from enum import Enum


class WriteMode(Enum):
    APPEND = "APPEND"  # Simply append rows each time
    TRUNCATE_INSERT = "TRUNCATE_INSERT"  # Fully load a table each time
    UPDATE_INSERT = "UPDATE_INSERT"  # Update and/or insert, but do not delete
    OVERWRITE_INSERT = "OVERWRITE_INSERT"  # Update and/or insert, but _do_ delete first
    # MERGE = "MERGE"  # Wait until postgres 15
    VIEW = "VIEW"  # Simply update view in place


class ModelType(Enum):
    TABLE = "TABLE"
    VIEW = "VIEW"



import types
import unittest
from unittest.mock import MagicMock
from bollhav.match import _parse_tag_expression, _module_matches


class TestParseTagExpression(unittest.TestCase):
    def test_single_or_group(self):
        self.assertEqual(_parse_tag_expression("[wee|x]"), [[["wee", "x"]]])

    def test_single_tag(self):
        self.assertEqual(_parse_tag_expression("[wee]"), [[["wee"]]])

    def test_and_group(self):
        self.assertEqual(_parse_tag_expression("[xyz&abc]"), [[["xyz"], ["abc"]]])

    def test_and_with_or_parens(self):
        self.assertEqual(_parse_tag_expression("[xyz&(c|e)]"), [[["xyz"], ["c", "e"]]])

    def test_multiple_groups(self):
        self.assertEqual(
            _parse_tag_expression("[wee|x],[xyz&(c|e)]"),
            [[["wee", "x"]], [["xyz"], ["c", "e"]]],
        )

    def test_no_brackets_raises(self):
        with self.assertRaises(ValueError):
            _parse_tag_expression("wee|x")

    def test_empty_raises(self):
        with self.assertRaises(ValueError):
            _parse_tag_expression("")


class TestModuleMatches(unittest.TestCase):
    def _make_module(self, tags: set[str]) -> types.ModuleType:
        module = MagicMock()
        model = MagicMock()
        model.tags = tags
        module.model = model
        return module

    def test_or_match(self):
        module = self._make_module({"wee"})
        self.assertTrue(_module_matches(module, _parse_tag_expression("[wee|x]")))

    def test_or_no_match(self):
        module = self._make_module({"abc"})
        self.assertFalse(_module_matches(module, _parse_tag_expression("[wee|x]")))

    def test_and_match(self):
        module = self._make_module({"xyz", "abc"})
        self.assertTrue(_module_matches(module, _parse_tag_expression("[xyz&abc]")))

    def test_and_missing_one(self):
        module = self._make_module({"xyz"})
        self.assertFalse(_module_matches(module, _parse_tag_expression("[xyz&abc]")))

    def test_and_with_or_parens_match(self):
        module = self._make_module({"xyz", "e"})
        self.assertTrue(_module_matches(module, _parse_tag_expression("[xyz&(c|e)]")))

    def test_and_with_or_parens_no_match(self):
        module = self._make_module({"xyz"})
        self.assertFalse(_module_matches(module, _parse_tag_expression("[xyz&(c|e)]")))

    def test_multiple_groups_first_matches(self):
        module = self._make_module({"wee"})
        self.assertTrue(
            _module_matches(module, _parse_tag_expression("[wee|x],[xyz&(c|e)]"))
        )

    def test_multiple_groups_second_matches(self):
        module = self._make_module({"xyz", "c"})
        self.assertTrue(
            _module_matches(module, _parse_tag_expression("[wee|x],[xyz&(c|e)]"))
        )

    def test_multiple_groups_none_match(self):
        module = self._make_module({"abc"})
        self.assertFalse(
            _module_matches(module, _parse_tag_expression("[wee|x],[xyz&(c|e)]"))
        )

    def test_no_model_attribute(self):
        module = MagicMock(spec=[])
        self.assertFalse(_module_matches(module, _parse_tag_expression("[wee]")))


if __name__ == "__main__":
    unittest.main()



from datetime import datetime, timedelta, timezone
from unittest.mock import MagicMock, patch
from typing import Any

import pytest

from bollhav.modes import WriteMode, ModelType
from bollhav.model import Model  # adjust import path as needed


UTC = timezone.utc


def utc_dt(year: int, month: int, day: int, hour: int = 0) -> datetime:
    return datetime(year, month, day, hour, tzinfo=UTC)


def make_column(name: str, unique: bool = False, sensitive: bool = False) -> MagicMock:
    col = MagicMock()
    col.name = name
    col.unique = unique
    col.sensitive = sensitive
    return col


def make_db() -> MagicMock:
    return MagicMock()


def minimal_model(**overrides: Any) -> Model:
    defaults: dict[str, Any] = {
        "name": "test_model",
        "source_entity": "source",
    }
    defaults.update(overrides)
    return Model(**defaults)


# --- __init__ validation ---


class TestInitValidation:
    def test_view_model_type_requires_view_write_mode(self) -> None:
        with pytest.raises(ValueError, match="ModelType.VIEW must use WriteMode.VIEW"):
            minimal_model(model_type=ModelType.VIEW, write_mode=WriteMode.APPEND)

    def test_table_model_type_rejects_view_write_mode(self) -> None:
        with pytest.raises(
            ValueError, match="ModelType.TABLE cannot use WriteMode.VIEW"
        ):
            minimal_model(model_type=ModelType.TABLE, write_mode=WriteMode.VIEW)

    def test_database_without_columns_raises(self) -> None:
        with pytest.raises(ValueError, match="columns must be set"):
            minimal_model(database=make_db(), columns=None)

    def test_columns_without_database_raises(self) -> None:
        cols = [make_column("id")]
        with pytest.raises(ValueError, match="database must be set"):
            minimal_model(database=None, columns=cols)

    def test_partitioned_by_unknown_column_raises(self) -> None:
        db = make_db()
        cols = [make_column("id")]
        with pytest.raises(
            ValueError, match="Partitioned_by references unknown column"
        ):
            minimal_model(database=db, columns=cols, partitioned_by="nonexistent")

    def test_partitioned_by_known_column_ok(self) -> None:
        db = make_db()
        cols = [make_column("ts")]
        m = minimal_model(database=db, columns=cols, partitioned_by="ts")
        assert m.partitioned_by == "ts"
        assert m.partitioned_by_index is True

    def test_partitioned_by_index_false_when_not_set(self) -> None:
        m = minimal_model()
        assert m.partitioned_by_index is False

    def test_begin_naive_tz_aware_raises(self) -> None:
        with pytest.raises(ValueError, match="begin must be UTC-aware"):
            minimal_model(begin=datetime(2024, 1, 1))

    def test_begin_non_utc_raises(self) -> None:
        non_utc = timezone(timedelta(hours=2))
        with pytest.raises(ValueError, match="begin must be UTC-aware"):
            minimal_model(begin=datetime(2024, 1, 1, tzinfo=non_utc))

    def test_end_naive_tz_aware_raises(self) -> None:
        with pytest.raises(ValueError, match="end must be UTC-aware"):
            minimal_model(end=datetime(2024, 1, 1))

    def test_begin_end_utc_ok(self) -> None:
        m = minimal_model(begin=utc_dt(2024, 1, 1), end=utc_dt(2024, 2, 1))
        assert m.begin == utc_dt(2024, 1, 1)
        assert m.end == utc_dt(2024, 2, 1)

    def test_tz_aware_false_skips_begin_end_check(self) -> None:
        m = minimal_model(tz_aware=False, begin=datetime(2024, 1, 1))
        assert m.begin == datetime(2024, 1, 1)

    def test_update_insert_without_unique_columns_raises(self) -> None:
        db = make_db()
        cols = [make_column("id", unique=False)]
        with pytest.raises(
            ValueError, match="requires at least one column with unique=True"
        ):
            minimal_model(database=db, columns=cols, write_mode=WriteMode.UPDATE_INSERT)

    def test_update_insert_with_unique_column_ok(self) -> None:
        db = make_db()
        cols = [make_column("id", unique=True)]
        m = minimal_model(database=db, columns=cols, write_mode=WriteMode.UPDATE_INSERT)
        assert len(m.unique_columns) == 1


# --- sensitive / unique_columns derivation ---


class TestDerivedAttributes:
    def test_sensitive_true_when_any_column_sensitive(self) -> None:
        db = make_db()
        cols = [make_column("id"), make_column("secret", sensitive=True)]
        m = minimal_model(database=db, columns=cols)
        assert m.sensitive is True

    def test_sensitive_false_when_no_sensitive_columns(self) -> None:
        db = make_db()
        cols = [make_column("id"), make_column("name")]
        m = minimal_model(database=db, columns=cols)
        assert m.sensitive is False

    def test_sensitive_false_without_columns(self) -> None:
        m = minimal_model()
        assert m.sensitive is False

    def test_unique_columns_collected(self) -> None:
        db = make_db()
        cols = [make_column("id", unique=True), make_column("name")]
        m = minimal_model(database=db, columns=cols)
        assert len(m.unique_columns) == 1
        assert m.unique_columns[0].name == "id"

    def test_unique_columns_empty_without_columns(self) -> None:
        m = minimal_model()
        assert m.unique_columns == []

    @patch("bollhav.model.infer_batch_size", return_value=100)
    def test_batch_size_inferred_from_cron(self, mock_infer: MagicMock) -> None:
        m = minimal_model(cron="0 * * * *")
        assert m.batch_size == 100
        mock_infer.assert_called_once_with("0 * * * *")

    def test_batch_size_none_without_cron(self) -> None:
        m = minimal_model()
        assert m.batch_size is None


# --- column sorting ---


class TestColumnSorting:
    def test_columns_sorted_via_column_sorting(self) -> None:
        db = make_db()
        cols = [make_column("b"), make_column("a")]
        # sorting fn reverses the list
        m = minimal_model(
            database=db,
            columns=cols,
            column_sorting=lambda names: list(reversed(names)),
        )
        assert [c.name for c in m.columns] == ["a", "b"]

    def test_no_sorting_when_column_sorting_is_none(self) -> None:
        db = make_db()
        cols = [make_column("b"), make_column("a")]
        m = minimal_model(database=db, columns=cols, column_sorting=None)
        assert [c.name for c in m.columns] == ["b", "a"]


# --- kwargs / extra ---


class TestKwargsExtra:
    def test_non_callable_kwargs_stored_in_extra(self) -> None:
        m = minimal_model(foo="bar", baz=42)
        assert m.extra == {"foo": "bar", "baz": 42}

    def test_callable_kwargs_resolved(self) -> None:
        fn = MagicMock(return_value="resolved")
        m = minimal_model(my_val=fn, static="s")
        assert m.extra["my_val"] == "resolved"


# --- __repr__ ---


class TestRepr:
    def test_repr_contains_name(self) -> None:
        m = minimal_model(name="my_model")
        assert "my_model" in repr(m)

    def test_repr_contains_source_entity(self) -> None:
        m = minimal_model(source_entity="my_source")
        assert "my_source" in repr(m)


# --- __eq__ ---


class TestEq:
    def test_equal_models(self) -> None:
        m1 = minimal_model()
        m2 = minimal_model()
        assert m1 == m2

    def test_unequal_models(self) -> None:
        m1 = minimal_model(name="a")
        m2 = minimal_model(name="b")
        assert m1 != m2

    def test_eq_with_non_model_returns_not_implemented(self) -> None:
        m = minimal_model()
        assert m.__eq__("not a model") == NotImplemented


# --- get_batch_intervals ---


class TestGetBatchIntervals:
    def _make_interval(self, since: datetime, until: datetime) -> MagicMock:
        interval = MagicMock()
        interval.since = since
        interval.until = until
        return interval

    @patch("bollhav.model.croniter")
    def test_returns_intervals_split_by_cron_ticks(
        self, mock_croniter: MagicMock
    ) -> None:
        since = utc_dt(2024, 1, 1, 0)
        tick1 = utc_dt(2024, 1, 1, 1)
        until = utc_dt(2024, 1, 1, 2)
        mock_it = MagicMock()
        mock_it.get_next.side_effect = [tick1, until]
        mock_croniter.return_value = mock_it
        m = minimal_model(cron="0 * * * *")
        interval = self._make_interval(since, until)
        result = m.get_batch_intervals(interval)
        assert len(result) == 2
        assert result[0].since == since
        assert result[0].until == tick1
        assert result[1].since == tick1
        assert result[1].until == until

    @patch("bollhav.model.croniter")
    def test_cron_override_used_when_provided(self, mock_croniter: MagicMock) -> None:
        since = utc_dt(2024, 1, 1)
        until = utc_dt(2024, 1, 2)
        mock_it = MagicMock()
        mock_it.get_next.side_effect = [until]
        mock_croniter.return_value = mock_it
        m = minimal_model(cron="0 * * * *")
        interval = self._make_interval(since, until)
        m.get_batch_intervals(interval, cron_override="0 0 * * *")
        mock_croniter.assert_called_with("0 0 * * *", since)

    @patch("bollhav.model.croniter")
    def test_no_ticks_before_until_returns_single_interval(
        self, mock_croniter: MagicMock
    ) -> None:
        since = utc_dt(2024, 1, 1)
        until = utc_dt(2024, 1, 2)
        mock_it = MagicMock()
        mock_it.get_next.side_effect = [until]
        mock_croniter.return_value = mock_it
        m = minimal_model(cron="0 * * * *")
        interval = self._make_interval(since, until)
        result = m.get_batch_intervals(interval)
        assert len(result) == 1
        assert result[0].since == since
        assert result[0].until == until

    @patch("bollhav.model.croniter")
    def test_tick_exactly_at_until_not_included(self, mock_croniter: MagicMock) -> None:
        since = utc_dt(2024, 1, 1)
        until = utc_dt(2024, 1, 1, 6)
        tick = utc_dt(2024, 1, 1, 6)
        mock_it = MagicMock()
        mock_it.get_next.side_effect = [tick]
        mock_croniter.return_value = mock_it
        m = minimal_model(cron="0 * * * *")
        interval = self._make_interval(since, until)
        result = m.get_batch_intervals(interval)
        assert len(result) == 1
        assert result[0].since == since



from datetime import datetime, timedelta
from typing import Callable
from icron import croniter
from bollhav.database import Database
from bollhav.intervals import TZInterval
from bollhav.postgres.columns import PostgresColumn
from bollhav.parquet.columns import ParquetColumn
from bollhav.modes import WriteMode, ModelType
from bollhav.batching import infer_batch_size
from bollhav.sorting import sort_columns


class Model:
    def __init__(
        self,
        name: str,
        source_entity: str,
        table: str = "",
        schema: str = "",
        database: Database | None = None,
        columns: list[PostgresColumn | ParquetColumn] | None = None,
        model_type: ModelType = ModelType.TABLE,
        write_mode: WriteMode = WriteMode.APPEND,
        tags: set[str] = set(),
        cron: str | None = None,
        enabled: bool = True,
        debug: bool = False,
        description: str | None = None,
        target_dsn_env_var: str | None = None,
        source_dsn_env_var: str | None = None,
        source_query: str | None = None,
        column_sorting: Callable | None = sort_columns,
        partitioned_by: str | None = None,
        partitioned_by_index: bool = False,
        begin: datetime | None = None,
        end: datetime | None = None,
        retries: int | None = None,
        lookback: int | None = None,
        name_add_to_tags: bool = True,
        schema_add_to_tags: bool = True,
        model_gets_all_tag: bool = True,
        tz_aware: bool = True,
        **kwargs,
    ):
        if model_type == ModelType.VIEW and write_mode != WriteMode.VIEW:
            raise ValueError("ModelType.VIEW must use WriteMode.VIEW")
        if model_type == ModelType.TABLE and write_mode == WriteMode.VIEW:
            raise ValueError("ModelType.TABLE cannot use WriteMode.VIEW")
        if database is not None and columns is None:
            raise ValueError("columns must be set when database is provided")
        if columns is not None and database is None:
            raise ValueError("database must be set when columns is provided")
        if partitioned_by and columns:
            col_names = {c.name for c in columns}
            if partitioned_by not in col_names:
                raise ValueError(
                    f"Partitioned_by references unknown column: {partitioned_by!r}"
                )
        if tz_aware:
            if begin is not None and (
                begin.tzinfo is None or begin.utcoffset() != timedelta(0)
            ):
                raise ValueError("begin must be UTC-aware")
            if end is not None and (
                end.tzinfo is None or end.utcoffset() != timedelta(0)
            ):
                raise ValueError("end must be UTC-aware")

        self.name = name
        self.source_entity = source_entity
        self.table = table
        self.schema = schema
        self.database = database
        self.columns = columns
        self.model_type = model_type
        self.write_mode = write_mode
        self.tags = set(tags) if tags else set()
        self.name_add_to_tags = name_add_to_tags
        if self.name_add_to_tags:
            tags.add(self.name)
        self.schema_add_to_tags = schema_add_to_tags
        if self.schema_add_to_tags:
            tags.add(self.schema)
        self.every_model_gets_all_tag = model_gets_all_tag
        if self.every_model_gets_all_tag:
            tags.add("all")
        self.cron = cron
        self.enabled = enabled
        self.debug = debug
        self.description = description
        self.target_dsn_env_var = target_dsn_env_var
        self.source_dsn_env_var = source_dsn_env_var
        self.source_query = source_query
        self.column_sorting = column_sorting
        self.partitioned_by = partitioned_by
        self.partitioned_by_index = partitioned_by is not None
        self.begin = begin
        self.end = end
        self.retries = retries
        self.lookback = lookback
        self.tz_aware = tz_aware
        self.batch_size = infer_batch_size(cron) if cron else None
        self.sensitive = (
            any(getattr(c, "sensitive", False) for c in columns) if columns else False
        )
        self.unique_columns = (
            [c for c in columns if getattr(c, "unique", False)] if columns else []
        )

        if write_mode == WriteMode.UPDATE_INSERT and not self.unique_columns:
            raise ValueError(
                "WriteMode.UPDATE_INSERT requires at least one column with unique=True"
            )

        if self.columns and self.column_sorting:
            col_names = [c.name for c in self.columns]
            sorted_names = self.column_sorting(col_names)
            name_to_col = {c.name: c for c in self.columns}
            self.columns = [name_to_col[n] for n in sorted_names]

        for key, val in kwargs.items():
            if callable(val):
                kwargs[key] = val(
                    **{k: v for k, v in kwargs.items() if not callable(v)}
                )
        self.extra = kwargs

    def __repr__(self) -> str:
        return (
            f"Model("
            f"name={self.name!r}, "
            f"source_entity={self.source_entity!r}, "
            f"table={self.table!r}, "
            f"schema={self.schema!r}, "
            f"database={self.database}, "
            f"columns={self.columns!r}, "
            f"model_type={self.model_type}, "
            f"write_mode={self.write_mode}, "
            f"tags={self.tags!r}, "
            f"cron={self.cron!r}, "
            f"enabled={self.enabled}, "
            f"debug={self.debug}, "
            f"description={self.description!r}, "
            f"source_dsn={self.source_dsn_env_var!r}, "
            f"source_query={self.source_query!r}, "
            f"partitioned_by={self.partitioned_by!r}, "
            f"begin={self.begin.isoformat() if self.begin else None!r}, "
            f"end={self.end.isoformat() if self.end else None!r}, "
            f"retries={self.retries!r}, "
            f"lookback={self.lookback!r}, "
            f"sensitive={self.sensitive}, "
            f"unique_columns={self.unique_columns!r}, "
            f"tz_aware={self.tz_aware}, "
            f"extra={self.extra!r})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Model):
            return NotImplemented
        return self.__dict__ == other.__dict__

    def get_batch_intervals(
        self, interval: TZInterval, cron_override: str | None = None
    ) -> list[TZInterval]:
        it = croniter(cron_override or self.cron, interval.since)
        intervals: list[TZInterval] = []
        current = interval.since
        while True:
            tick = it.get_next(datetime)
            if tick >= interval.until:
                break
            intervals.append(TZInterval(current, tick))
            current = tick
        if current < interval.until:
            intervals.append(TZInterval(current, interval.until))
        return intervals



import importlib
import re
from pathlib import Path
import sys
import types
from typing import Callable


def _parse_tag_expression(expr: str) -> list[list[str]]:
    groups = re.findall(r"\[([^\]]+)\]", expr)
    if not groups:
        raise ValueError(
            f"Invalid tag expression: {expr}. Groups must be wrapped in []."
        )

    result: list[list[str]] = []
    for group in groups:
        if "&" in group:
            and_terms = group.split("&")
            or_sets: list[list[str]] = []
            for term in and_terms:
                term = term.strip()
                paren_match = re.match(r"^\(([^)]+)\)$", term)
                if paren_match:
                    or_sets.append([t.strip() for t in paren_match.group(1).split("|")])
                else:
                    or_sets.append([term])
            result.append(or_sets)
        else:
            result.append([[t.strip() for t in group.split("|")]])

    return result


def _module_matches(
    module: types.ModuleType,
    parsed: list,
) -> bool:
    model = getattr(module, "model", None)
    if model is None:
        return False
    module_tags: set[str] = getattr(model, "tags", set())
    for group in parsed:
        if all(any(tag in module_tags for tag in or_set) for or_set in group):
            return True
    return False


def match_execute_functions(
    folder: str = "src/models",
    tags: str | None = None,
) -> list[Callable]:
    """
    Discover and return execute functions from Python modules in a folder recursively,
    filtered by a tag expression.

    TAG EXPRESSION SYNTAX
    =====================

    Tags are matched against the `tags` attribute on each module's `model` object.
    The expression must be passed as a string using the syntax described below.

    GROUPS
    ------
    A group is wrapped in square brackets: [expression]
    Multiple groups are separated by commas: [group1],[group2]

    A model is included if it matches ANY group (outer OR).

        [wee],[xyz]         # match if model has "wee" OR "xyz"

    OR WITHIN A GROUP
    -----------------
    Use | to express OR within a group.
    A model matches the group if it has ANY of the tags.

        [wee|x]             # match if model has "wee" OR "x"

    AND WITHIN A GROUP
    ------------------
    Use & to express AND within a group.
    A model matches the group if it has ALL of the terms.

        [xyz&abc]           # match if model has "xyz" AND "abc"

    AND WITH OR SUB-EXPRESSION
    --------------------------
    Use parentheses to group OR terms within an AND expression.
    Only one level of parentheses is supported.

        [xyz&(c|e)]         # match if model has "xyz" AND ("c" OR "e")

    COMBINING
    ---------
    Groups can be combined freely with commas.

        [wee|x],[xyz&(c|e)] # match if:
                            #   model has "wee" OR "x"
                            #   OR model has "xyz" AND ("c" OR "e")

    ENVIRONMENT VARIABLE
    --------------------
    The expression is typically sourced from an environment variable:

        export TAGS="[wee|x],[xyz&(c|e)]"

    LIMITATIONS
    -----------
    - Square brackets are required around every group
    - Only one level of parentheses is supported
    - & and | cannot be mixed at the top level without parentheses

    Args:
        folder: Path to the folder containing model modules. Defaults to "src/models".
        tags: Tag filter expression string. Must be provided.

    Returns:
        List of callable execute functions from matched modules.

    Raises:
        ValueError: If tags is not provided or the expression is invalid.
    """
    if not tags:
        raise ValueError("tags must be a non-empty expression.")

    parsed = _parse_tag_expression(tags)
    print(f"Filtering execute functions to those with tags: {tags}")

    folder_path = Path(folder)
    parent_dir = str(folder_path.parent)
    added_to_path = False
    if parent_dir not in sys.path:
        sys.path.insert(0, parent_dir)
        added_to_path = True

    try:
        execute_functions = []
        for file in folder_path.rglob("*.py"):
            module_spec = importlib.util.spec_from_file_location(
                name=file.stem, location=file
            )
            module = importlib.util.module_from_spec(spec=module_spec)
            module_spec.loader.exec_module(module)
            if _module_matches(module, parsed):
                execute_functions.append(module.execute)
        return execute_functions
    finally:
        if added_to_path:
            sys.path.remove(parent_dir)



from dataclasses import dataclass, fields
from bollhav.database import DatabaseColumn
from enum import Enum


class ParquetType(Enum):
    BOOLEAN = "BOOLEAN"
    INT32 = "INT32"
    INT64 = "INT64"
    INT96 = "INT96"  # deprecated, legacy only
    FLOAT = "FLOAT"
    DOUBLE = "DOUBLE"
    BYTE_ARRAY = "BYTE_ARRAY"
    FIXED_LEN_BYTE_ARRAY = "FIXED_LEN_BYTE_ARRAY"


@dataclass
class ParquetColumn(DatabaseColumn):
    data_type: ParquetType = ParquetType.BYTE_ARRAY
    length: int | None = None
    precision: int | None = None
    scale: int | None = None

    def __post_init__(self) -> None:
        if self.data_type == ParquetType.FIXED_LEN_BYTE_ARRAY and self.length is None:
            raise ValueError(
                f"Column {self.name!r}: FIXED_LEN_BYTE_ARRAY requires length"
            )

        decimal_types = {
            ParquetType.INT32,
            ParquetType.INT64,
            ParquetType.FIXED_LEN_BYTE_ARRAY,
        }
        has_decimal_annotation = self.precision is not None or self.scale is not None

        if has_decimal_annotation and self.data_type not in decimal_types:
            raise ValueError(
                f"Column {self.name!r}: precision/scale (DECIMAL annotation) only valid "
                f"on INT32, INT64, or FIXED_LEN_BYTE_ARRAY, not {self.data_type.value}"
            )
        if has_decimal_annotation and not (
            self.precision is not None and self.scale is not None
        ):
            raise ValueError(
                f"Column {self.name!r}: precision and scale must both be set for DECIMAL annotation"
            )

    def __repr__(self) -> str:
        base_parts = [
            f"{f.name}={getattr(self, f.name)!r}" for f in fields(DatabaseColumn)
        ]
        parts = base_parts + [f"data_type={self.data_type}"]
        if self.length is not None:
            parts.append(f"length={self.length}")
        if self.precision is not None:
            parts.append(f"precision={self.precision}")
        if self.scale is not None:
            parts.append(f"scale={self.scale}")
        return f"ParquetColumn({', '.join(parts)})"



from bollhav.model import Model
from bollhav.modes import ModelType, WriteMode
from bollhav.batching import BatchSize
from bollhav.database import Database, DatabaseColumn
from bollhav.intervals import TZInterval
from bollhav.match import match_execute_functions

__all__ = [
    "Model",
    "ModelType",
    "WriteMode",
    "BatchSize",
    "Database",
    "DatabaseColumn",
    "TZInterval",
    "match_execute_functions",
]
